from dataclasses import dataclass
from typing import Union

from colors_accessibility.utils import check_if_values_are_positive, check_if_correct_hex_value, check_if_values_are_numeric, \
    check_if_correct_hsl_values
from colors_accessibility.utils.decorators import validate_parameters
from colors_accessibility.utils.errors import ExceededMaximumValueError, IncorrectColorSpaceError, IncorrectInputLengthError, \
    IncorrectHexLengthError
from colors_accessibility.utils.registry import Registry
from colors_accessibility.data.configuration_data import CONFIGURATION_DATA


@dataclass
class Validator:
    color_space_validators = Registry()

    @staticmethod
    @validate_parameters
    def validate_color_space(color_space: str) -> bool:
        if color_space.lower() not in CONFIGURATION_DATA.get('implemented'):
            raise IncorrectColorSpaceError(color_space, CONFIGURATION_DATA.get('implemented'))
        return True

    @staticmethod
    @validate_parameters
    def validate_input_length(color_space: str, values: Union[str, dict, list, tuple]) -> bool:
        if isinstance(values, str):
            values = [values]
        values_length = len(values)
        if values_length != CONFIGURATION_DATA.get('lengths').get(color_space):
            raise IncorrectInputLengthError(color_space, values_length)
        return True

    @staticmethod
    @color_space_validators.register('rgb')
    def validate_rgb_values(values: Union[list, tuple]) -> bool:
        if not values:
            raise IncorrectInputLengthError('rgb', 0)
        check_if_values_are_numeric(values)
        check_if_values_are_positive(values)
        exceeded_values = []
        for value in values:
            if value > CONFIGURATION_DATA.get('maximum_values').get('rgb'):
                exceeded_values.append(value)
        if exceeded_values:
            raise ExceededMaximumValueError(exceeded_values, CONFIGURATION_DATA.get('maximum_values').get('rgb'), 'rgb')
        return True

    @staticmethod
    @color_space_validators.register('hex')
    def validate_hex_values(values: Union[list, str]) -> bool:
        if type(values) == list:
            values = ''.join(values)
        value = values.replace('#', '')
        if len(value) not in [3, 6]:
            raise IncorrectHexLengthError(value)
        check_if_correct_hex_value(values)
        return True

    @staticmethod
    @color_space_validators.register('hsl')
    def validate_hsl_values(values: list) -> bool:
        check_if_correct_hsl_values(values)
        return True

    def validate_values(self, color_space: str, values: Union[list, tuple]) -> bool:
        self.validate_input_length(color_space, values)
        self.color_space_validators.functions.get(color_space)(values)
        return True

    def validate(self, color_space: str, values: Union[str, list, tuple]) -> bool:
        self.validate_color_space(color_space)
        self.validate_values(color_space, values)
        return True



from dataclasses import dataclass


@dataclass
class ColorRepresentation:
    normalized: list = None
    extended: list = None
    dictionary: dict = None


@dataclass
class ColorRepresentations:
    rgb: ColorRepresentation = None
    hex: ColorRepresentation = None
    hsl: ColorRepresentation = None



class Registry:
    def __init__(self):
        self.functions = {}

    def register(self, name: str):
        def wrapped(function):
            self.functions[name] = function
            return function
        return wrapped



from dataclasses import dataclass

from .color import Color
from .utils import update_non_empty_values
from .data.configuration_data import CONFIGURATION_DATA


@dataclass
class AccessibilityProcessor:
    foreground_color: Color
    background_color: Color

    def __post_init__(
            self
    ):
        self.foreground_color.color_values = self.foreground_color.to_rgb()
        self.foreground_color.color_space = 'rgb'
        self.background_color.color_values = self.background_color.to_rgb()
        self.background_color.color_space = 'rgb'

    @staticmethod
    def get_luminance_values(
            foreground_color: Color,
            background_color: Color
    ) -> tuple:
        return foreground_color.calculate_relative_luminance(), background_color.calculate_relative_luminance()

    @staticmethod
    def calculate_contrast(
            foreground_luminance: float,
            background_luminance: float
    ) -> float:
        maximum_color_relative_luminance = max(foreground_luminance, background_luminance)
        minimum_color_relative_luminance = min(foreground_luminance, background_luminance)
        color_contrast = (maximum_color_relative_luminance + 0.05) / (minimum_color_relative_luminance + 0.05)
        return color_contrast

    def calculate(
            self
    ):
        foreground_luminance, background_luminance = self.get_luminance_values(self.foreground_color,
                                                                               self.background_color)
        contrast = self.calculate_contrast(foreground_luminance, background_luminance)
        return contrast

    def get_parameter_signature(
            self,
            parameter: str,
            color_to_change: str,
            value: float = 0.01
    ) -> int:
        value *= 5
        original_contrast = self.calculate_contrast(self.foreground_color.calculate_relative_luminance(),
                                                    self.background_color.calculate_relative_luminance())
        parameter_index = 1 if parameter == 'saturation' else 2
        color_to_change_values = self.foreground_color.to_hsl() \
            if color_to_change == 'foreground' \
            else self.background_color.to_hsl()
        color_to_change_values[parameter_index] += value
        color_to_change_values[parameter_index] = max(0, min(color_to_change_values[parameter_index], 1))
        changed_color = Color('hsl', color_to_change_values)
        if color_to_change == 'foreground':
            contrast = self.calculate_contrast(changed_color.calculate_relative_luminance(),
                                               self.background_color.calculate_relative_luminance())
        else:
            contrast = self.calculate_contrast(self.foreground_color.calculate_relative_luminance(),
                                               changed_color.calculate_relative_luminance())
        return 1 if contrast >= original_contrast else -1

    def get_parameters_signatures(
            self,
            parameter: str,
            color_to_change: str,
            value: float = 0.01
    ) -> dict:
        parameters_signatures = {}
        if parameter == 'saturation' or 'both':
            saturation_signature = self.get_parameter_signature('saturation', color_to_change, value)
            parameters_signatures.update({
                1: saturation_signature
            })
        if parameter == 'lightness' or 'both':
            lightness_signature = self.get_parameter_signature('lightness', color_to_change, value)
            parameters_signatures.update({
                2: lightness_signature
            })
        return parameters_signatures

    @staticmethod
    def increment_color_hsl_values(
            color: Color,
            parameter: str,
            parameters_signatures: dict,
            value: float = 0.01
    ) -> Color:
        hsl_values = color.to_hsl()
        indexes_to_update = []
        if parameter in ['saturation', 'both']:
            indexes_to_update.append(1)
        if parameter in ['lightness', 'both']:
            indexes_to_update.append(2)
        for index in indexes_to_update:
            value *= parameters_signatures.get(index)
            if index == 1 and (CONFIGURATION_DATA.get('hsl_check_values_ranges').get('saturation').get('max')
                               <= hsl_values[index] or hsl_values[index]
                               <= CONFIGURATION_DATA.get('hsl_check_values_ranges').get('saturation').get('min')):
                value = 0
            if index == 2 and (CONFIGURATION_DATA.get('hsl_check_values_ranges').get('lightness').get('max')
                               <= hsl_values[index] or hsl_values[index]
                               <= CONFIGURATION_DATA.get('hsl_check_values_ranges').get('lightness').get('min')):
                value = 0
            hsl_values[index] += value
        return Color('hsl', hsl_values)

    @staticmethod
    def update_wcag_requirements(
            contrast: float,
            foreground_color_values: list = None,
            background_color_values: list = None
    ) -> dict:
        if foreground_color_values:
            foreground_color_values = [round(value, 2) for value in foreground_color_values]
        if background_color_values:
            background_color_values = [round(value, 2) for value in background_color_values]
        values = [color_values for color_values in [foreground_color_values, background_color_values] if color_values]
        wcag_requirements_fulfilled = {}
        if contrast >= 3:
            wcag_requirements_fulfilled.update({
                'large_aa': values
            })
        if contrast >= 4.5:
            wcag_requirements_fulfilled.update({
                'normal_aa': values
            })
            wcag_requirements_fulfilled.update({
                'large_aaa': values
            })
        if contrast >= 7:
            wcag_requirements_fulfilled.update({
                'normal_aaa': values
            })
        return wcag_requirements_fulfilled

    @staticmethod
    def check_borders_reached(
            color: Color,
            borders_reached: dict
    ) -> dict:
        hsl_color_values = color.to_hsl()
        if CONFIGURATION_DATA.get('hsl_check_values_ranges').get('saturation').get('max') <= hsl_color_values[1] \
                or hsl_color_values[1] <= CONFIGURATION_DATA.get('hsl_check_values_ranges').get('saturation').get(
                'min'):
            borders_reached.update({
                'saturation': True
            })
        if CONFIGURATION_DATA.get('hsl_check_values_ranges').get('lightness').get('max') <= hsl_color_values[2] \
                or hsl_color_values[2] <= CONFIGURATION_DATA.get('hsl_check_values_ranges').get('lightness').get('min'):
            borders_reached.update({
                'lightness': True
            })
        return borders_reached

    @staticmethod
    def check_borders_reached_for_both(
            color: Color,
            color_type: str,
            borders_reached: dict
    ) -> dict:
        hsl_color_values = color.to_hsl()
        if CONFIGURATION_DATA.get('hsl_check_values_ranges').get('saturation').get('max') <= hsl_color_values[1] \
                or hsl_color_values[1] <= CONFIGURATION_DATA.get('hsl_check_values_ranges').get('saturation').get(
                'min'):
            color_key = 'saturation_foreground' if color_type == 'foreground' else 'saturation_background'
            borders_reached.update({
                color_key: True
            })
        if CONFIGURATION_DATA.get('hsl_check_values_ranges').get('lightness').get('max') <= hsl_color_values[2] \
                or hsl_color_values[2] <= CONFIGURATION_DATA.get('hsl_check_values_ranges').get('lightness').get('min'):
            color_key = 'lightness_foreground' if color_type == 'foreground' else 'lightness_background'
            borders_reached.update({
                color_key: True
            })
        return borders_reached

    def update_wcag_colors_for_single_color(
            self,
            changed_color: Color,
            constant_color: Color,
            wcag_colors: dict
    ) -> dict:
        contrast = self.calculate_contrast(
            changed_color.calculate_relative_luminance(),
            constant_color.calculate_relative_luminance()
        )
        fulfilled_requirements = self.update_wcag_requirements(contrast, changed_color.to_hsl())
        wcag_compliant_colors = update_non_empty_values(wcag_colors, fulfilled_requirements)
        return wcag_compliant_colors

    def update_wcag_colors_for_both_colors(
            self,
            foreground_color: Color,
            background_color: Color,
            wcag_colors: dict
    ) -> dict:
        contrast = self.calculate_contrast(
            foreground_color.calculate_relative_luminance(),
            background_color.calculate_relative_luminance()
        )
        fulfilled_requirements = self.update_wcag_requirements(
            contrast, foreground_color.to_hsl(), background_color.to_hsl())
        wcag_compliant_colors = update_non_empty_values(wcag_colors, fulfilled_requirements)
        return wcag_compliant_colors

    def find_wcag_compliant_single_color_change(
            self,
            color_to_change: str,
            parameter: str,
            value: float = 0.01
    ) -> dict:
        wcag_compliant_colors = {}
        borders_reached = {
            'saturation': True if parameter == 'lightness' else False,
            'lightness': True if parameter == 'saturation' else False
        }
        parameters_signatures = self.get_parameters_signatures(parameter, color_to_change, value)
        background_color = Color(self.background_color.original_color_space, self.background_color.original_values)
        background_color.set_to_hsl()
        changed_color = self.foreground_color \
            if color_to_change == 'foreground' \
            else background_color
        constant_color = background_color \
            if color_to_change == 'foreground' \
            else self.foreground_color
        wcag_compliant_colors = self.update_wcag_colors_for_single_color(
            changed_color, constant_color, wcag_compliant_colors
        )
        while True:
            changed_color = self.increment_color_hsl_values(changed_color, parameter, parameters_signatures)
            borders_reached = self.check_borders_reached(changed_color, borders_reached)
            if borders_reached.get('saturation') and borders_reached.get('lightness'):
                break
            wcag_compliant_colors = self.update_wcag_colors_for_single_color(
                changed_color, constant_color, wcag_compliant_colors
            )
        return wcag_compliant_colors

    def find_wcag_compliant_both_colors_changes(
            self,
            parameter: str,
            value: float = 0.01
    ) -> dict:
        wcag_compliant_colors = {}
        borders_reached = {
            'saturation_foreground': True if parameter == 'lightness' else False,
            'lightness_foreground': True if parameter == 'saturation' else False,
            'saturation_background': True if parameter == 'lightness' else False,
            'lightness_background': True if parameter == 'saturation' else False
        }
        # Get parameter signatures
        foreground_parameters_signatures = self.get_parameters_signatures(parameter, 'foreground', value)
        background_parameters_signatures = {key: -1 * value for key, value in foreground_parameters_signatures.items()}
        # Set foreground and background colors
        changed_foreground_color = self.foreground_color
        changed_background_color = Color(self.background_color.original_color_space,
                                         self.background_color.original_values)
        changed_background_color.set_to_hsl()
        # Check if original colors are already compliant, if so add them to the dictionary
        wcag_compliant_colors = self.update_wcag_colors_for_both_colors(
            changed_foreground_color, changed_background_color, wcag_compliant_colors)
        # Run while loop until color pair is compliant, break it once both colors parameters reach their bounds and
        # return the dictionary with found color pairs
        while True:
            changed_foreground_color = self.increment_color_hsl_values(
                changed_foreground_color, parameter, foreground_parameters_signatures)
            changed_background_color = self.increment_color_hsl_values(
                changed_background_color, parameter, background_parameters_signatures)
            borders_reached = self.check_borders_reached_for_both(
                changed_foreground_color, 'foreground', borders_reached)
            borders_reached = self.check_borders_reached_for_both(
                changed_background_color, 'background', borders_reached)
            if borders_reached.get('saturation_foreground') and borders_reached.get('lightness_foreground') \
                    and borders_reached.get('saturation_background') and borders_reached.get('lightness_background'):
                break
            wcag_compliant_colors = self.update_wcag_colors_for_both_colors(
                changed_foreground_color, changed_background_color, wcag_compliant_colors)
        return wcag_compliant_colors

    def get_representation_single_color(self, color_to_change: str, parameter: str, value: float = 0.01) -> dict:
        colors_data = self.find_wcag_compliant_single_color_change(color_to_change, parameter, value)
        if not colors_data:
            return {}
        representation = {
            key: [
                Color('hsl', values)
                for values
                in values_list
            ]
            for key, values_list
            in colors_data.items()
        }
        return representation

    def get_representation_multiple_color(self, parameter: str, value: float = 0.01) -> dict:
        colors_data = self.find_wcag_compliant_both_colors_changes(parameter, value)
        representation = {
            key: [
                Color('hsl', values)
                for values
                in values_list
            ]
            for key, values_list
            in colors_data.items()
        }
        return representation

    def get_all_wcag_compliant_color(self, value: float = 0.01) -> dict:
        wcag_compliant_colors = {
            'saturation': {
                'foreground': None,
                'background': None,
            },
            'lightness': {
                'foreground': None,
                'background': None,
            }
        }
        for parameter in ['saturation', 'lightness']:
            wcag_compliant_colors[parameter]['foreground'] = self.get_representation_single_color(
                'foreground', parameter, value
            )
            wcag_compliant_colors[parameter]['background'] = self.get_representation_single_color(
                'background', parameter, value
            )
            wcag_compliant_colors[parameter]['both'] = self.get_representation_multiple_color(
                parameter, value
            )
        return wcag_compliant_colors



CONFIGURATION_DATA = {
    'implemented': ['rgb', 'hex', 'hsl'],
    'maximum_values': {
        'rgb': 255
    },
    'lengths': {
        'rgb': 3,
        'hex': 1,
        'hsl': 3
    },
    'hsl_check_values_ranges': {
        'saturation': {
            'min': 0.05,
            'max': 0.95
        },
        'lightness': {
            'min': 0.05,
            'max': 0.95
        }
    }
}



from decimal import Decimal
from typing import Any, Union
from string import hexdigits

from colors_accessibility.utils.errors import IncorrectArgumentsTypingError, IncorrectHexValueError, IncorrectHslValueError, \
    IncorrectValueTypeError, NegativeNumberError


def check_if_values_are_positive(
        values: Any
) -> bool:
    incorrect_values = []
    for value in values:
        if value < 0:
            incorrect_values.append(value)
    if incorrect_values:
        raise NegativeNumberError(incorrect_values)
    return True


def check_if_correct_hex_value(
        value: str
) -> bool:
    value = value.replace('#', '')
    incorrect_values = [
        character
        for character
        in value
        if character not in hexdigits
    ]
    if incorrect_values:
        raise IncorrectHexValueError(value, incorrect_values)
    return True


def check_if_correct_hsl_values(
        values: list
) -> bool:
    check_if_values_are_positive(values)
    hue, saturation, lightness = values
    incorrect_values = {}
    if not 0 <= hue <= 360:
        incorrect_values['hue'] = hue
    if not 0 <= saturation <= 1:
        incorrect_values['saturation'] = saturation
    if not 0 <= lightness <= 1:
        incorrect_values['lightness'] = lightness
    if incorrect_values:
        raise IncorrectHslValueError(incorrect_values)
    return True


def check_if_values_are_numeric(
        values: Union[list, tuple]
) -> bool:
    incorrect_values = {}
    for value in values:
        if not type(value) in [int, float, Decimal, str]:
            incorrect_values[value] = type(value).__name__
        if type(value) == str and not value.isnumeric():
            try:
                float(value)
            except ValueError:
                incorrect_values[value] = type(value).__name__
    if incorrect_values:
        raise IncorrectValueTypeError(incorrect_values)
    return True


def normalize_value(
        value: Any,
        factor: int = 255,
        decimal_places: int = 20
) -> float:
    if value >= 1:
        value /= factor
    return round(value, decimal_places)


def normalize_hsl_values(
        values: list
) -> list:
    if 0 < values[0] < 1:
        values[0] *= 360
    if 1 < values[1] <= 100:
        values[1] /= 100
    if 1 < values[2] <= 100:
        values[2] /= 100
    return values


def normalize_values(
        values: Union[list, tuple],
        factor: int = 255,
        decimal_places: int = 20
) -> list:
    return [normalize_value(value, factor, decimal_places) for value in values]


def convert_integer_to_hex(
        number: int,
) -> str:
    hex_number = '%02x' % number
    return hex_number


def expand_values(
        values: list,
        factor: int = 255
) -> list:
    return [
        int(value * factor)
        if 0 < value <= 1
        else value
        for value
        in values
    ]


def expand_hsl_values(
        values: list
) -> list:
    if 0 < values[1] <= 1:
        values[1] *= 100
    if 0 < values[2] <= 1:
        values[2] *= 100
    return values


def associate_input_args_with_expected_types(
        args_values: tuple,
        argument_names: list
) -> dict:
    return {
        argument: value
        for value, argument
        in zip(args_values, argument_names)
    }


def update_non_empty_values(
        dictionary_to_update: dict,
        updates_dict: dict
) -> dict:
    for key, value in updates_dict.items():
        if not dictionary_to_update.get(key):
            dictionary_to_update[key] = value
    return dictionary_to_update



from dataclasses import dataclass, field
from typing import Union

from colors_accessibility.utils import normalize_hsl_values, normalize_values, expand_values, convert_integer_to_hex, expand_hsl_values
from colors_accessibility.utils.classes import ColorRepresentation, ColorRepresentations
from colors_accessibility.utils.errors import TypeCastingError
from colors_accessibility.utils.registry import Registry
from colors_accessibility.validator import Validator
from colors_accessibility.data.configuration_data import CONFIGURATION_DATA


@dataclass
class Color:
    color_space: str
    color_values: Union[str, dict, list, tuple]
    implemented_color_spaces: list = field(default_factory=lambda: CONFIGURATION_DATA.get('implemented'))

    validator: Validator = Validator()

    conversion_registry: Registry = Registry()
    processing_registry: Registry = Registry()
    setter_registry: Registry = Registry()
    to_rgb_registry: Registry = Registry()
    to_hex_registry: Registry = Registry()
    to_hsl_registry: Registry = Registry()
    to_dict_registry: Registry = Registry()

    def __str__(self):
        return f"Color(color_space='{self.color_space}', color_values='{self.color_values}')"

    def __repr__(self):
        return f"Color(color_space='{self.color_space}', color_values='{self.color_values}')"

    def cast_to_type(self):
        types_mapping = {
            'rgb': float,
            'hsl': float
        }
        incorrect_values = []
        correct_values = []
        for value in self.color_values:
            try:
                correct_values.append(types_mapping.get(self.color_space)(value))
            except (TypeError, ValueError):
                incorrect_values.append(value)
        if incorrect_values:
            raise TypeCastingError(incorrect_values, str(types_mapping.get(self.color_space)))
        self.color_values = correct_values

    def format_hex(self):
        if type(self.color_values) == list:
            self.color_values = ''.join(self.color_values)
        color_value = self.color_values.replace('#', '')
        if len(color_value) == 3:
            self.color_values = f'#{color_value[0] * 2}{color_value[1] * 2}{color_value[2] * 2}'

    @processing_registry.register('rgb')
    def process_rgb(self):
        self.cast_to_type()
        self.validator.validate('rgb', self.color_values)
        self.color_values = normalize_values(self.color_values, 255)

    @processing_registry.register('hex')
    def process_hex(self):
        self.validator.validate('hex', self.color_values)
        self.format_hex()

    @processing_registry.register('hsl')
    def process_hsl(self):
        self.cast_to_type()
        self.color_values = normalize_hsl_values(self.color_values)
        self.validator.validate('hsl', self.color_values)
        if 0 < self.color_values[0] < 1:
            self.color_values[0] = int(self.color_values[0] * 360)

    def process(self, color_space: str):
        if type(self.color_values) == dict:
            self.color_values = list(self.color_values.values())
        self.processing_registry.functions.get(color_space)(self)

    def __post_init__(self):
        self.original_values = self.color_values
        self.original_color_space = self.color_space
        self.color_space = self.color_space.lower()
        self.process(self.color_space)

    def process_hsl_values(self, number: int) -> float:
        k = (number + self.color_values[0] / 30) % 12
        a = self.color_values[1] * min([self.color_values[2], 1 - self.color_values[2]])
        return self.color_values[2] - a * max([-1, min([k - 3, 9 - k, 1])])

    @to_hex_registry.register('rgb')
    def rgb_to_hex(self):
        color_values = expand_values(self.color_values)
        hex_representation = '#'
        for index, value in enumerate(color_values):
            value = convert_integer_to_hex(int(value))
            hex_representation += value
        return hex_representation

    @to_hex_registry.register('hsl')
    def hsl_to_hex(self):
        original_values = self.color_values
        rgb_representation = self.hsl_to_rgb()
        self.color_values = rgb_representation
        hex_representation = self.rgb_to_hex()
        self.color_values = original_values
        return hex_representation

    @to_rgb_registry.register('hex')
    def hex_to_rgb(self):
        hex_value = self.color_values.replace('#', '')
        return [
            int(hex_value[index: index + 2], 16)
            for index
            in range(0, len(hex_value), 2)
        ]

    @to_rgb_registry.register('hsl')
    def hsl_to_rgb(self) -> list:
        rgb_representation = [
            round(self.process_hsl_values(0) * 255),
            round(self.process_hsl_values(8) * 255),
            round(self.process_hsl_values(4) * 255)
        ]
        return rgb_representation

    @to_hsl_registry.register('rgb')
    def rgb_to_hsl(self, rounding_factor: int = 2) -> list:
        red_value, green_value, blue_value = expand_values(self.color_values)
        red_value /= 255
        green_value /= 255
        blue_value /= 255
        max_value = max(red_value, green_value, blue_value)
        min_value = min(red_value, green_value, blue_value)
        lightness = (max_value + min_value) / 2
        if max_value == min_value:
            saturation = 0
            hue = 0
        else:
            saturation = lightness < 0.5 and (max_value - min_value) / (max_value + min_value) or (
                    max_value - min_value) / (2 - max_value - min_value)
            if red_value == max_value:
                hue = (green_value - blue_value) / (max_value - min_value)
            elif green_value == max_value:
                hue = 2 + (blue_value - red_value) / (max_value - min_value)
            else:
                hue = 4 + (red_value - green_value) / (max_value - min_value)
        hue *= 60
        if hue < 0:
            hue += 360
        return [round(hue, 0), round(saturation, rounding_factor), round(lightness, rounding_factor)]

    @to_hsl_registry.register('hex')
    def hex_to_hsl(self) -> list:
        original_values = self.color_values
        rgb_representation = self.hex_to_rgb()
        self.color_values = rgb_representation
        hsl_representation = self.rgb_to_hsl()
        self.color_values = original_values
        return hsl_representation

    @conversion_registry.register('hex')
    def to_hex(self):
        if self.color_space == 'hex':
            return self.color_values
        return self.to_hex_registry.functions.get(self.color_space)(self)

    @conversion_registry.register('rgb')
    def to_rgb(self):
        if self.color_space == 'rgb':
            return self.color_values
        return normalize_values(self.to_rgb_registry.functions.get(self.color_space)(self), 255)

    @conversion_registry.register('hsl')
    def to_hsl(self):
        if self.color_space == 'hsl':
            return self.color_values
        return self.to_hsl_registry.functions.get(self.color_space)(self)

    @setter_registry.register('rgb')
    def set_to_rgb(self):
        self.color_values = self.to_rgb()
        self.color_space = 'rgb'

    @setter_registry.register('hex')
    def set_to_hex(self):
        self.color_values = self.to_hex()
        self.color_space = 'hex'

    @setter_registry.register('hsl')
    def set_to_hsl(self):
        self.color_values = self.to_hsl()
        self.color_space = 'hsl'

    def set_to(self, color_space: str):
        color_space = color_space.lower()
        self.validator.validate_color_space(color_space)
        self.setter_registry.functions.get(color_space)(self)

    def calculate_relative_luminance(self, new_estimated_threshold: bool = False) -> float:
        rgb_representation = self.color_values if self.color_space == 'rgb' else self.to_rgb()
        threshold = 0.03928 if not new_estimated_threshold else 0.04045
        calculated_color_weights = [
            color_value / 12.92
            if color_value <= threshold
            else ((color_value + 0.055) / 1.055) ** 2.4
            for color_value
            in rgb_representation
        ]
        relative_luminance = 0.2126 * calculated_color_weights[0] \
                             + 0.7152 * calculated_color_weights[1] \
                             + 0.0722 * calculated_color_weights[2]
        return relative_luminance

    def convert(self, color_space: str):
        color_values = getattr(self, f'to_{color_space}')()
        self.color_values = color_values
        self.color_space = color_space

    @to_dict_registry.register('rgb')
    def rgb_to_dict(self):
        return {
            'red': self.color_values[0],
            'green': self.color_values[1],
            'blue': self.color_values[2]
        }

    @to_dict_registry.register('hex')
    def hex_to_dict(self):
        return {
            'hex': self.color_values
        }

    @to_dict_registry.register('hsl')
    def hsl_to_dict(self):
        return {
            'hue': self.color_values[0],
            'saturation': self.color_values[1],
            'lightness': self.color_values[2]
        }

    def to_dict(self):
        return self.to_dict_registry.functions.get(self.color_space)(self)

    def get_representations(self, color_space: str = None):
        original_color_space = self.color_space
        original_color_values = self.color_values
        color_representations = {}
        if not color_space:
            color_space = self.color_space
        if color_space in ['rgb', 'all']:
            self.convert('rgb')
            normalized = self.color_values if original_color_space != 'rgb' else original_color_values
            extended = expand_values(normalized)
            dictionary = self.to_dict()
            color_representations['rgb'] = ColorRepresentation(normalized, extended, dictionary)
        if color_space in ['hex', 'all']:
            self.convert('hex')
            normalized = self.color_values if original_color_space != 'hex' else original_color_values
            extended = normalized
            dictionary = self.to_dict()
            color_representations['hex'] = ColorRepresentation(normalized, extended, dictionary)
        if color_space in ['hsl', 'all']:
            self.convert('hsl')
            normalized = self.color_values if original_color_space != 'hsl' else original_color_values
            extended = expand_hsl_values(normalized)
            dictionary = self.to_dict()
            color_representations['hsl'] = ColorRepresentation(normalized, extended, dictionary)
        return ColorRepresentations(**color_representations)



from decimal import Decimal
from typing import Union, List

from colors_accessibility.data.configuration_data import CONFIGURATION_DATA


class ExceededMaximumValueError(Exception):
    def __init__(
            self,
            values: List[Union[int, float, Decimal]],
            maximum_value: int,
            color_space: str
    ):
        self.values = ",\n\t".join([
            f'{value}'
            for value
            in values
        ])
        self.maximum_value = maximum_value
        self.color_space = color_space

    def __str__(self):
        return f'The maximum value for values in "{self.color_space}" color space is ' \
               f'"{CONFIGURATION_DATA.get("maximum_values").get(self.color_space)}". ' \
               f'Incorrect input values:\n\t{self.values}.'


class IncorrectArgumentsTypingError(Exception):
    def __init__(self, expected_typing_mapping: dict, incorrect_typing_mapping: dict):
        self.expected_typing_mapping = expected_typing_mapping
        self.incorrect_typing_mapping = incorrect_typing_mapping

    def __str__(self):
        arguments_message = ',\n\t'.join([
            f'{key}: {self.incorrect_typing_mapping.get(key)} -> {self.expected_typing_mapping.get(key)}'
            for key
            in self.incorrect_typing_mapping.keys()
        ])
        return f'Some of the provided arguments have incorrect typings. Incorrect arguments:\n\t{arguments_message}.'


class IncorrectColorSpaceError(Exception):
    def __init__(self, color_space_name: str, implemented_color_spaces: list):
        self.color_space_name = color_space_name
        self.implemented_color_spaces = self.values = ",\n\t".join([
            f'{value}'
            for value
            in implemented_color_spaces
        ])

    def __str__(self):
        return f'The color name space must be one of: {self.implemented_color_spaces}. ' \
               f'Input color space: "{self.color_space_name}".'


class IncorrectHexLengthError(Exception):
    def __init__(self, hex_value: str):
        self.hex_value = hex_value

    def __str__(self):
        return f'The hex value "{self.hex_value}" must be of length 3 or 6. (# sign is optional and does not count ' \
               f'toward length of value).Provided value: "{self.hex_value}" has length of {len(self.hex_value)}.'


class IncorrectHexValueError(Exception):
    def __init__(self, hex_value: str, incorrect_values: list):
        self.hex_value = hex_value
        self.incorrect_values = incorrect_values

    def __str__(self):
        return f'The hex value "{self.hex_value}" must be a valid hex value.\n' \
               f'Provided value: {self.hex_value} is not a valid hex value.\n' \
               f'Incorrect characters in input: {", ".join(self.incorrect_values)}.'


class IncorrectHslValueError(Exception):
    def __init__(self, incorrect_values_mapping: dict):
        self.incorrect_values_mapping = incorrect_values_mapping

    def __str__(self):
        message = ',\n\t'.join([
            f'{key}: {self.incorrect_values_mapping.get(key)}'
            for key
            in self.incorrect_values_mapping.keys()
        ])
        return f'The provided HSL values must be within the range of hue: [0-360], saturation: [0-1]  as % or [0-100]' \
               f' as absolute value, lightness: [0-1] as % or [0-100] as absolute value.\n' \
               f'Incorrect values:\n\t{message}.'


class IncorrectInputLengthError(Exception):
    def __init__(self, color_space: str, input_length: int):
        self.color_space = color_space
        self.input_length = input_length

    def __str__(self):
        return f'The input length for "{self.color_space}" color space must be' \
               f' {CONFIGURATION_DATA.get("lengths").get(self.color_space)}. Input length: "{self.input_length}".'


class IncorrectValueTypeError(Exception):
    def __init__(self, values_with_incorrect_types: dict):
        self.values_with_incorrect_types = ",\n\t".join([
            f'{key}: {value}'
            for key, value
            in values_with_incorrect_types.items()
        ])

    def __str__(self):
        return f'The input value must of type "decimal", "float", "int" or "str" and be convertable to numeric.\n\t' \
               f'Incorrect values with corresponding types are:\n\t{self.values_with_incorrect_types}.'


class NegativeNumberError(Exception):
    def __init__(self, values: List[Union[int, float, Decimal]]):
        self.values = ",\n\t".join([
            f'{value}'
            for value
            in values
        ])

    def __str__(self):
        return f'The input values must be positive numerics. Incorrect input values:\n\t{self.values}.'


class TypeCastingError(Exception):
    def __init__(self, values: list, expected_type: str):
        self.values = values
        self.expected_type = expected_type

    def __str__(self):
        message = ',\n\t'.join([
            f'{value}: {type(value)}'
            for value
            in self.values
        ])
        return f'Some of the provided arguments cannot be casted to the type "{self.expected_type}". ' \
               f'Incorrect values:\n\t{message}.'



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="colors_accessibility",
    version="0.0.5",
    author="Piotr Hryniewicz",
    author_email="phryniewicz.dev@gmail.com",
    description="Package to process and change colors between color spaces and to tweak input colors to meet WCAG 2.1 "
                "accessibility standards.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/phryniewicz/colors-accessibility",
    project_urls={
        "Bug Tracker": "https://github.com/phryniewicz/colors-accessibility/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)



from functools import wraps
from colors_accessibility.utils import associate_input_args_with_expected_types
from typing import get_args, get_type_hints
from colors_accessibility.utils.errors import IncorrectArgumentsTypingError


def validate_parameters(function: callable):
    @wraps(function)
    def wrapper(*args, **kwargs):
        arguments_typing = get_type_hints(function)
        arguments_typing = {
            key: get_args(value) or [value]
            for key, value
            in arguments_typing.items()
        }
        args_typing = associate_input_args_with_expected_types(args, list(arguments_typing.keys()))
        input_typings = {**kwargs, **args_typing}
        incorrect_arguments_typing = {
            key: type(value)
            for key, value
            in input_typings.items()
            if key in arguments_typing and type(value) not in arguments_typing.get(key)
        }
        correct_arguments_typing = {
            key: (value[0] if type(value) == list else value)
            for key, value
            in arguments_typing.items()
            if key in incorrect_arguments_typing
        }
        if incorrect_arguments_typing:
            raise IncorrectArgumentsTypingError(correct_arguments_typing, incorrect_arguments_typing)
        return function(*args, **kwargs)
    return wrapper
